unit uMain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, LCLType, Controls, Graphics, Dialogs,
  SynEdit, Menus, SynEditHighlighter, SynHighlighterAny,
  SynHighlighterPas, SynHighlighterCpp, SynMacroRecorder, ComCtrls, Buttons, uOptions,
  uProcess, ActnList, StrUtils, SynHighlighterJava, SynHighlighterPHP,
  synhighlighterunixshellscript, SynEditKeyCmds, ExtCtrls, StdCtrls, SynEditTypes,
  Process, IniFiles, SynHighlighterHTML, SynHighlighterPerl,
  SynHighlighterXML, SynHighlighterCss, SynHighlighterTeX, SynHighlighterSQL,
  SynHighlighterPython, Printers, PrintersDlgs, uInfo;
type

  TVersion = packed record
    case Integer of
      0: (AsInt: Cardinal;);
      1: (AsArray: Array[0..3] of Byte;);
      2: (Major: Byte; Minor: Byte; Release: Byte; Build: Byte;);
  end;

  { TSettings }

  TSettings = class(TIniFile)
  public
    function loadBool(id: String): Boolean;
    procedure saveBool(id: String; value: Boolean);
    function loadString(id: String): String;
    procedure saveString(id: String; value: String);
    function loadInt(id: String): Integer;
    procedure saveInt(id: String; value: Integer);
    function loadFont(id: String): TFont;
    procedure saveFont(id: String; value: TFont);
    function loadColor(id: String): TColor;
    procedure saveColor(id: String; value: TColor);
  end;

  { TLanguageList }

  TLanguageList = class(TStringList)
  private
    FOwner: TComponent;
  public
    function GetCompileString(language: String): String;
    function GetRunString(language: String): String;
    function GetExtensionString(language: String): String;
    function GetHighlighter(language: String): TSynCustomHighlighter;
    procedure SetCompileString(language: String; s: String);
    procedure SetRunString(language: String; s: String);
    procedure SetExtensionString(language: String; s: String);
    procedure SetHighlighter(language: String; s: TSynCustomHighlighter);
    function LanguageCount: Integer;
    function GetLanguage(index: Integer): String;
    procedure AddLanguage(language: String; CompileString: String; RunString: String; ExtensionString: String; Highlighter: TSynCustomHighlighter);
    procedure ExportToFile(FileName: String);
    procedure ImportFromFile(FileName: String);
    constructor Create(Owner: TComponent);
  end;

  TToolButtons = Array of TSpeedButton;

  { TfrmMain }

  TfrmMain = class(TForm)
    aClose: TAction;
    aCodingOptions: TAction;
    aCompile: TAction;
    aPlayback: TAction;
    aRecordMacro: TAction;
    aToggleBreakpoint: TAction;
    aDebug: TAction;
    aNewWindow: TAction;
    aToggleSearch: TAction;
    aRun: TAction;
    aInfo: TAction;
    aLineNumbers: TAction;
    aFont: TAction;
    aFindNext: TAction;
    aFindPrev: TAction;
    aReplace: TAction;
    aJumpTo: TAction;
    aFind: TAction;
    aRedo: TAction;
    aCut: TAction;
    aCopy: TAction;
    aPaste: TAction;
    aDelete: TAction;
    aSelectAll: TAction;
    aUndo: TAction;
    aPrint: TAction;
    aSaveAs: TAction;
    aSave: TAction;
    aNew: TAction;
    aOpen: TAction;
    alMain: TActionList;
    cbMatchCase: TCheckBox;
    cbWholeWord: TCheckBox;
    cbRegExp: TCheckBox;
    edSearch: TEdit;
    fdMain: TFontDialog;
    ilMain: TImageList;
    MenuItem35: TMenuItem;
    MenuItem36: TMenuItem;
    mmMain: TMainMenu;
    MenuItem1: TMenuItem;
    MenuItem10: TMenuItem;
    MenuItem11: TMenuItem;
    MenuItem12: TMenuItem;
    MenuItem13: TMenuItem;
    MenuItem14: TMenuItem;
    MenuItem15: TMenuItem;
    MenuItem16: TMenuItem;
    MenuItem17: TMenuItem;
    MenuItem18: TMenuItem;
    MenuItem19: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem20: TMenuItem;
    MenuItem21: TMenuItem;
    MenuItem22: TMenuItem;
    MenuItem23: TMenuItem;
    MenuItem24: TMenuItem;
    MenuItem25: TMenuItem;
    MenuItem26: TMenuItem;
    MenuItem27: TMenuItem;
    MenuItem28: TMenuItem;
    MenuItem29: TMenuItem;
    MenuItem3: TMenuItem;
    MenuItem30: TMenuItem;
    MenuItem31: TMenuItem;
    MenuItem32: TMenuItem;
    MenuItem33: TMenuItem;
    MenuItem34: TMenuItem;
    MenuItem4: TMenuItem;
    MenuItem5: TMenuItem;
    MenuItem6: TMenuItem;
    MenuItem7: TMenuItem;
    MenuItem8: TMenuItem;
    MenuItem9: TMenuItem;
    odMain: TOpenDialog;
    pnlSearchButtons: TPanel;
    pnlSearch: TPanel;
    pmLanguages: TPopupMenu;
    pdMain: TPrintDialog;
    sdMain: TSaveDialog;
    hlCPP: TSynCppSyn;
    sbToggleSearch: TSpeedButton;
    hlJava: TSynJavaSyn;
    sbMain: TStatusBar;
    hlCSS: TSynCssSyn;
    hlHTML: TSynHTMLSyn;
    mrMain: TSynMacroRecorder;
    synMain: TSynEdit;
    hlFPC: TSynFreePascalSyn;
    hlPHP: TSynPHPSyn;
    hlShell: TSynUNIXShellScriptSyn;
    hlPerl: TSynPerlSyn;
    hlPython: TSynPythonSyn;
    hlSQL: TSynSQLSyn;
    hlTEX: TSynTeXSyn;
    hlXML: TSynXMLSyn;
    tbButtons: TToolBar;
    procedure aCloseExecute(Sender: TObject);
    procedure aCodingOptionsExecute(Sender: TObject);
    procedure aCompileExecute(Sender: TObject);
    procedure aCopyExecute(Sender: TObject);
    procedure aCutExecute(Sender: TObject);
    procedure aDebugExecute(Sender: TObject);
    procedure aDeleteExecute(Sender: TObject);
    procedure aFindExecute(Sender: TObject);
    procedure aFindNextExecute(Sender: TObject);
    procedure aFindPrevExecute(Sender: TObject);
    procedure aFontExecute(Sender: TObject);
    procedure aInfoExecute(Sender: TObject);
    procedure aJumpToExecute(Sender: TObject);
    procedure aLineNumbersExecute(Sender: TObject);
    procedure aNewExecute(Sender: TObject);
    procedure aNewWindowExecute(Sender: TObject);
    procedure aOpenExecute(Sender: TObject);
    procedure aPasteExecute(Sender: TObject);
    procedure aPlaybackExecute(Sender: TObject);
    procedure aPrintExecute(Sender: TObject);
    procedure aRecordMacroExecute(Sender: TObject);
    procedure aRedoExecute(Sender: TObject);
    procedure aReplaceExecute(Sender: TObject);
    procedure aRunExecute(Sender: TObject);
    procedure aSaveAsExecute(Sender: TObject);
    procedure aSaveExecute(Sender: TObject);
    procedure aSelectAllExecute(Sender: TObject);
    procedure aToggleBreakpointExecute(Sender: TObject);
    procedure aToggleSearchExecute(Sender: TObject);
    procedure aUndoExecute(Sender: TObject);
    procedure edSearchChange(Sender: TObject);
    procedure edSearchKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure edSearchKeyPress(Sender: TObject; var Key: char);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure pmLanguagesPopup(Sender: TObject);
    procedure synMainGutterClick(Sender: TObject; X, Y, Line: integer;
      mark: TSynEditMark);
    procedure synMainPaint(Sender: TObject; ACanvas: TCanvas);
    procedure synMainProcessUserCommand(Sender: TObject;
      var Command: TSynEditorCommand; var AChar: TUTF8Char; Data: pointer);
    procedure synMainSpecialLineColors(Sender: TObject; Line: integer;
      var Special: boolean; var FG, BG: TColor);
  private
    { private declarations }
    FBreakPoints: Array of Integer;
    FFilename: String;
    FLanguage: String;
    FLanguageList: TLanguageList;
    FMultipleWindows: Boolean;
    FToolButtons: TToolButtons;
    FActualLine: Integer;

    procedure UpdateFileName(filename: String);
    function AskForSave: Boolean;
    function SaveFile(SaveAs: Boolean): Boolean;
    procedure OpenFile(filename: String);
    function GetLanguageFromFileName(FileName: String): String;
    procedure pmLanguagesClick(Sender: TObject);

    procedure LoadSettings;
    procedure SaveSettings;

    procedure RunFile(dodebug: Boolean);
    procedure CompileFile;
    procedure RunCommand(cmd: String; dodebug: Boolean = false);

  public
    { public declarations }
    procedure CreateToolButtons;
    procedure ExchangeActionItems(i, j: Integer);
    procedure SetLanguage(AValue: String);
    procedure ChangeBreakpoint(line: Integer);
    procedure DelBreakPoint(line: Integer);
    procedure AddBreakPoint(line: Integer);
    function BreakPointCount: Integer;
    function GetBreakPoint(index: Integer): Integer;
    function IndexOfBreakPoint(line: Integer): Integer;

    property Language: String read FLanguage write SetLanguage;
    property LanguageList: TLanguageList read FLanguageList write FLanguageList;
    property MultipleWindows: Boolean read FMultipleWindows write FMultipleWindows;
    property ToolButtons: TToolButtons read FToolButtons write FToolButtons;
    property FileName: String read FFileName;
    property ActualLine: Integer read FActualLine write FActualLine;
  end; 

var
  frmMain: TfrmMain;
  SettingsPath: String;
  Version: TVersion;
  TitleCaption: String;

const
  Title = 'Inheredit';
  C_BASE_COMMAND=1337;

implementation

function TfrmMain.GetLanguageFromFileName(FileName: String): String;
var ext, lang: String;
  i: Integer;
begin
  result := '';
  ext := ExtractFileExt(FileName)+'/';
  if ext = '/' then
    exit;
  ext := '/'+Copy(ext, 2, length(ext));
  for i := 0 to FLanguageList.LanguageCount-1 do
  begin
    lang := FLanguageList.GetLanguage(i);
    if pos(ext, FLanguageList.GetExtensionString(lang))>0 then
    begin
      result := lang;
      break;
    end;
  end;
end;

procedure TfrmMain.pmLanguagesClick(Sender: TObject);
begin
  SetLanguage(FLanguageList.GetLanguage((Sender as TMenuItem).Tag));
end;

procedure TfrmMain.LoadSettings;
var cfg: TSettings; s: String; i, j: Integer; k: Integer; f: TFont;
begin
  cfg := TSettings.Create(SettingsPath+'settings.cfg');
  try
    //Load the font
    f := cfg.loadFont('synMain.Font');
    if Assigned(f) then
      synMain.Font:=f;

    //Load the language settings
    FLanguageList.ImportFromFile(SettingsPath+'languages.cfg');

    //Load the Action list
    for i := 0 to alMain.ActionCount-1 do
    begin
      s :=cfg.loadString('Actions['+IntToStr(i)+']');
      k :=cfg.loadInt('Actions['+IntToStr(i)+'].Tag');
      if alMain.Actions[i].Name <> s then
      begin
        for j := i+1 to alMain.ActionCount-1 do
        begin
          if alMain.Actions[j].Name = s then
          begin
            ExchangeActionItems(i,j);
            break;
          end;
        end;
      end;
      alMain.Actions[i].Tag:=k;
    end;
    CreateToolButtons;

    //Load line number data
    aLineNumbers.Checked:= not cfg.loadBool('LineNumbers');
    aLineNumbers.Execute;

    //Load new window data
    FMultipleWindows := cfg.loadBool('MultipleWindows');
    frmOptions.cbMultibleWindows.Checked:=FMultipleWindows;

    //Load borland style editor
    if cfg.loadBool('BorlandStyle') then
      synMain.Options := synMain.Options + [eoScrollPastEol]
    else
      synMain.Options := synMain.Options - [eoScrollPastEol];
    frmOptions.cbBorlandStyle.Checked:=(eoScrollPastEol in synMain.Options);

    //Load tab size data
    k := cfg.loadInt('TabSize');
    if k = 0 then
      k := 2;
    synMain.TabWidth:=k;
    synMain.BlockIndent:=k;
    frmOptions.seTabSize.Value:=k;

    //Load Width and Height of Window
    i := cfg.LoadInt('Width');
    j := cfg.LoadInt('Height');
    if i <> 0 then
      Width := i;
    if j <> 0 then
      Height := j;

    //Load Debug Path
    frmOptions.edDebug.Text := cfg.loadString('DebugPath');
    {for i := 0 to ComponentCount-1 do
    begin
      if Components[i] is TSynCustomHighLighter then
        TSynCustomHighLighter(Components[i]).SaveToFile(SettingsPath+Components[i].Name+'.inc');
    end;}

    //Load macros
    try
      mrMain.LoadFromFile(SettingsPath+'macro.cfg');
    except
    end;
  finally
    cfg.Free;
  end;
end;

procedure TfrmMain.SaveSettings;
var cfg: TSettings; i: Integer;
begin
  cfg := TSettings.Create(SettingsPath+'settings.cfg');
  try
    //Save the font
    cfg.saveFont('synMain.Font', synMain.Font);

    //Save the language settings
    FLanguageList.ExportToFile(SettingsPath+'languages.cfg');

    //Save the Action list
    for i := 0 to alMain.ActionCount-1 do
    begin
      cfg.saveString('Actions['+IntToStr(i)+']', alMain.Actions[i].Name);
      cfg.saveInt('Actions['+IntToStr(i)+'].Tag', alMain.Actions[i].Tag);
    end;

    //Save line number data
    cfg.saveBool('LineNumbers', aLineNumbers.Checked);

    //Save new window data
    cfg.saveBool('MultipleWindows', frmOptions.cbMultibleWindows.Checked);

    //Save borland style editor
    cfg.saveBool('BorlandStyle', frmOptions.cbBorlandStyle.Checked);

    //Save tab size data
    cfg.saveInt('TabSize', frmOptions.seTabSize.Value);

    //Save Width and height of window
    cfg.saveInt('Width', Width);
    cfg.saveInt('Height', Height);

    //Save debug path
    cfg.saveString('DebugPath', frmOptions.edDebug.Text);

    //Save macros
    mrMain.SaveToFile(SettingsPath+'macro.cfg');
  finally
    cfg.Free;
  end;
end;

procedure TfrmMain.CreateToolButtons;
var i, c: Integer;
begin
  for i := 0 to length(FToolButtons)-1 do
  begin
    if Assigned(FToolButtons[i]) then
      FreeAndNil(FToolButtons[i]);
  end;

  SetLength(FToolButtons, alMain.ActionCount);
  c:=0;
  for i := 0 to alMain.ActionCount-1 do
  begin
    FToolButtons[i] := TSpeedButton.Create(tbButtons);
    with FToolButtons[i] do
    begin
      Width := 32;
      Height := 32;
      Left := c*Width;
      Parent:=tbButtons;
      Hint :=TAction(alMain.Actions[i]).Caption;
      ShowHint := true;
      Flat:=true;
      ParentShowHint:=true;
      Action :=TAction(alMain.Actions[i]);
      ShowCaption := false;
      Visible := alMain.Actions[i].Tag = 1;
      if Visible then
        inc(c);
    end;
  end;
end;

procedure TfrmMain.ExchangeActionItems(i, j: Integer);
var a: TAction; ts1, ts2: String;
begin
  a := TAction.Create(alMain);
  try
    a.Assign(alMain.Actions[i]);
    alMain.Actions[i].Assign(alMain.Actions[j]);
    alMain.Actions[j].Assign(a);
    ts1 := alMain.Actions[i].Name;
    ts2 := alMain.Actions[j].Name;
    alMain.Actions[i].Name := 'wuppdi'+IntToStr($DEADBEEF-Random($DEADBEEF));
    alMain.Actions[j].Name := ts1;
    alMain.Actions[i].Name := ts2;
  finally
    a.Free;
  end;
end;

{ TfrmMain }

procedure TfrmMain.FormCreate(Sender: TObject);
var i: Integer; a: Integer;
begin
  Version.Major:=0;
  Version.Minor:=4;
  Version.Release:=2;
  Version.Build:=5;

  SettingsPath := GetEnvironmentVariableUTF8('HOME')+'/.'+AnsiLowerCase(Title)+'/';
  TitleCaption := Title+' '+IntToStr(version.AsArray[0]);
  for i := 1 to 2 do
    TitleCaption := TitleCaption + '.'+IntToStr(version.AsArray[i]);
  TitleCaption := TitleCaption + '-'+IntToStr(version.Build);
  Caption := titlecaption;
  FLanguageList := TLanguageList.Create(self);
  for i := 0 to alMain.ActionCount-1 do
  begin
    a := synMain.Keystrokes.FindShortcut(TAction(alMain.Actions[i]).ShortCut);
    if a >= 0 then
      synMain.Keystrokes.Delete(a);
    with synMain.Keystrokes.Add do
    begin
      Command:=C_BASE_COMMAND+i;
      ShortCut:=TAction(alMain.Actions[i]).ShortCut;
    end;
  end;
  CreateToolButtons;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  FLanguageList.Free;
end;

procedure TfrmMain.FormShow(Sender: TObject);
var i: Integer;
begin
  LoadSettings;

  for i := 1 to ParamCount do
  begin
    OpenFile(ParamStr(i));
  end;
end;

procedure TfrmMain.pmLanguagesPopup(Sender: TObject);
var a: TMenuItem;
  i: Integer;
begin
  pmLanguages.Items.Clear;
  for i := 0 to FLanguageList.LanguageCount-1 do
  begin
    a := TMenuItem.Create(pmLanguages);
    a.Caption:=FLanguageList.GetLanguage(i);
    a.Tag:=i;
    a.OnClick:=@pmLanguagesClick;
    if FLanguage = a.Caption then
      a.Checked:=true;
    pmLanguages.Items.Add(a);
  end;
end;

procedure TfrmMain.synMainGutterClick(Sender: TObject; X, Y, Line: integer;
  mark: TSynEditMark);
begin
  ChangeBreakPoint(line);
  synMain.Repaint;
end;

procedure TfrmMain.synMainPaint(Sender: TObject; ACanvas: TCanvas);
begin
  sbMain.Panels.Items[0].Text:=IntToStr(synMain.CaretY)+': '+IntToStr(synMain.CaretX);
  sbMain.Panels.Items[1].Text:=BoolToStr(synMain.Modified, 'Changed', '');
  sbMain.Panels.Items[2].Text:=FLanguage;
  if FFilename = '' then
    sbMain.Panels.Items[3].Text:='Unsaved'
  else
    sbMain.Panels.Items[3].Text:=FFileName;
end;

procedure TfrmMain.synMainProcessUserCommand(Sender: TObject;
  var Command: TSynEditorCommand; var AChar: TUTF8Char; Data: pointer);
begin
  if (Command >= C_BASE_COMMAND) and (Command < C_BASE_COMMAND+alMain.ActionCount) then
    TAction(alMain.Actions[Command-C_BASE_COMMAND]).Execute;
end;

procedure TfrmMain.synMainSpecialLineColors(Sender: TObject; Line: integer;
  var Special: boolean; var FG, BG: TColor);
begin
  Special := true;
  if line mod 2 = 0 then
    BG := $FFF5F5
  else
    BG := clNone;

  if IndexOfBreakPoint(line) <> -1 then
    BG := clRed;

  if FActualLine = line then
    BG := clYellow;
end;

procedure TfrmMain.UpdateFileName(filename: String);
begin
  FFileName := FileName;
  if FileName <> '' then
    Caption := ExtractFileName(FileName) + ' - ' + TitleCaption
  else
    Caption := TitleCaption;
  Language := GetLanguageFromFileName(FFilename);
end;

function TfrmMain.AskForSave: Boolean;
begin
  result := true;
  if synMain.Modified then
  begin
    case Application.MessageBox('There is unsaved text! Save first?', 'Question', MB_ICONWARNING or MB_YESNOCANCEL) of
    IDYES: begin
      result := SaveFile(false);
    end;
    IDNO: begin
      result := true;
    end;
    IDCANCEL: begin
      result := false;
    end;
    end;
  end;
end;

function TfrmMain.SaveFile(SaveAs: Boolean): Boolean;
begin
  result := false;
  if FFilename = '' then
  begin
    SaveAs := true;
  end;

  if SaveAs then
  begin
    if sdMain.Execute then
    begin
      try
        synMain.Lines.SaveToFile(sdMain.FileName);
      except
        on e: Exception do
        begin
          ShowMessage('Error saving file '+sdMain.FileName+'! '+e.Message);
          exit;
        end;
      end;
      UpdateFileName(sdMain.FileName);
      synMain.Modified := false;
    end else
      result := false;
  end else
  begin
    try
      synMain.Lines.SaveToFile(FFileName);
    except
      on e: Exception do
      begin
        ShowMessage('Error saving file '+FFileName+'! '+e.Message);
        exit;
      end;
    end;

    synMain.Modified := false;
  end;
  result := true;
end;

procedure TfrmMain.OpenFile(Filename: String);
begin
  if (FMultipleWindows) and (not ((not synMain.Modified) and (synMain.Text=''))) then
  begin
    with TProcess.Create(nil) do
    begin
      CommandLine:=ParamStr(0)+' "'+FileName+'"';
      Execute;
    end;
  end
  else if AskForSave then
  begin
    try
      //ShowMessage(Filename);
      synMain.Lines.LoadFromFile(FileName);
    except
      on e: Exception do
      begin
        ShowMessage('Error opening file '+FFileName+'! '+e.Message);
        exit;
      end;
    end;
    UpdateFileName(Filename);
  end;
end;

procedure TfrmMain.SetLanguage(AValue: String);
begin
  FLanguage := AValue;
  synMain.Highlighter := FLanguageList.GetHighlighter(FLanguage);
  if Assigned(synMain.Highlighter) then
  begin
    if FileExists(SettingsPath+synMain.Highlighter.Name+'.inc') then
    begin
      try
        synMain.Highlighter.LoadFromFile(SettingsPath+synMain.Highlighter.Name+'.inc')
      except
        on e: Exception do
        begin
          ShowMessage('Error opening file '+SettingsPath+synMain.Highlighter.Name+'.inc! '+e.Message);
          exit;
        end;
      end;
    end else begin
      try
        synMain.Highlighter.SaveToFile(SettingsPath+synMain.Highlighter.Name+'.inc')
      except
      end;
    end;
  end;
end;

procedure TfrmMain.RunCommand(cmd: String; dodebug: Boolean = false);
begin
  cmd := StringReplace(cmd, '%p', ExcludeTrailingPathDelimiter(ExtractFilePath(FFileName)), [rfReplaceAll]);
  cmd := StringReplace(cmd, '%f', ExtractFileName(FFileName), [rfReplaceAll]);
  cmd := StringReplace(cmd, '%n', ChangeFileExt(ExtractFileName(FFileName), ''), [rfReplaceAll]);
  with TfrmProcess.Create(self) do
  begin
//    parent := self;
    Debug := dodebug;
    Output := not dodebug;
    DebugBin := frmOptions.edDebug.Text;
    Command:=cmd;
    WorkDir := ExtractFilePath(FFileName);
    Show;
  end;
end;

procedure TfrmMain.ChangeBreakpoint(line: Integer);
begin
  if IndexOfBreakpoint(line) = -1 then
  begin
    AddBreakPoint(line);
  end else begin
    DelBreakPoint(line);
  end;
end;

procedure TfrmMain.DelBreakPoint(line: Integer);
var i, j: Integer;
begin
  i := IndexOfBreakPoint(line);
  if i <> -1 then
  begin
    for j := Length(FBreakPoints)-2 downto i do
    begin
      FBreakPoints[j] := FBreakPoints[j+1];
    end;
    SetLength(FBreakPoints, Length(FBreakPoints)-1);
  end;
end;

procedure TfrmMain.AddBreakPoint(line: Integer);
begin
  if IndexOfBreakPoint(line) = -1 then
  begin
    SetLength(FBreakPoints, Length(FBreakPoints)+1);
    FBreakPoints[High(FBreakPoints)] := line;
  end;
end;

function TfrmMain.BreakPointCount: Integer;
begin
  result := Length(FBreakPoints);
end;

function TfrmMain.GetBreakPoint(index: Integer): Integer;
begin
  result := FBreakPoints[index];
end;

function TfrmMain.IndexOfBreakPoint(line: Integer): Integer;
var i: Integer;
begin
  result := -1;
  for i := 0 to Length(FBreakPoints)-1 do
  begin
    if FBreakPoints[i] = line then
    begin
      result := i;
      exit;
    end;
  end;
end;

procedure TfrmMain.RunFile(dodebug: Boolean);
var s: String;
begin
  s := FLanguageList.GetRunString(FLanguage);
  if s <> '' then
  begin
    RunCommand(s, dodebug);
  end;
end;

procedure TfrmMain.CompileFile;
var s: String;
begin
  s := FLanguageList.GetCompileString(FLanguage);
  if s <> '' then
  begin
    RunCommand(s, false);
  end;
end;

procedure TfrmMain.aNewExecute(Sender: TObject);
begin
  if AskForSave then
  begin
    synMain.Lines.Clear;
    synMain.Modified:=false;
    UpdateFileName('');
  end;
end;

procedure TfrmMain.aNewWindowExecute(Sender: TObject);
begin
  with TProcess.Create(nil) do
  begin
    CommandLine:=ParamStr(0);
    Execute;
  end;
end;

procedure TfrmMain.aCloseExecute(Sender: TObject);
begin
  Close;
end;

procedure TfrmMain.aCodingOptionsExecute(Sender: TObject);
begin
  if frmOptions.ShowModal = mrOK then
  begin
    SaveSettings;
    LoadSettings;
  end else begin
    LoadSettings;
  end;
end;

procedure TfrmMain.aCompileExecute(Sender: TObject);
begin
  if (FFileName <> '') or (SaveFile(true)) then
  begin
    if synMain.Modified then
      SaveFile(false);
    CompileFile;
  end;
end;

procedure TfrmMain.aCopyExecute(Sender: TObject);
begin
  synMain.CopyToClipboard;
end;

procedure TfrmMain.aCutExecute(Sender: TObject);
begin
  synMain.CutToClipboard;
end;

procedure TfrmMain.aDebugExecute(Sender: TObject);
begin
  if (FFileName <> '') or (SaveFile(true)) then
  begin
    if synMain.Modified then
      SaveFile(false);
    RunFile(true);
  end;
end;

procedure TfrmMain.aDeleteExecute(Sender: TObject);
begin
  synMain.SelText:='';
end;

procedure TfrmMain.aFindExecute(Sender: TObject);
begin
  aToggleSearch.Execute;
end;

procedure TfrmMain.aFindNextExecute(Sender: TObject);
var i: Integer;
begin
  //TODO!
  for i := 0 to mrMain.EventCount-1 do
    ShowMessage(mrMain.Events[i].AsString);
end;

procedure TfrmMain.aFindPrevExecute(Sender: TObject);
begin
  //TODO!
end;

procedure TfrmMain.aFontExecute(Sender: TObject);
begin
  if fdMain.Execute then
  begin
    synMain.Font.Assign(fdMain.Font);
  end;
end;

procedure TfrmMain.aInfoExecute(Sender: TObject);
begin
  //TODO: Show better Info
  //ShowMessage('by inherited aka. Nikolai Wyderka');
  frmInfo.Show;
end;

procedure TfrmMain.aJumpToExecute(Sender: TObject);
var s: String;
begin
  s := '';
  if InputQuery('Line Number', 'Which line number?' , s) then
  begin
    synMain.CaretY:=StrToIntDef(s, 0);
    synMain.CaretX:=0;
  end;
end;

procedure TfrmMain.aLineNumbersExecute(Sender: TObject);
begin
  aLineNumbers.Checked := not aLineNumbers.Checked;
  synMain.Gutter.Visible:=MenuItem26.Checked;
end;

procedure TfrmMain.aOpenExecute(Sender: TObject);
begin
  if odMain.Execute then
  begin
    openfile(odMain.Filename);
  end;
end;

procedure TfrmMain.aPasteExecute(Sender: TObject);
begin
  synMain.PasteFromClipboard;
end;

procedure TfrmMain.aPlaybackExecute(Sender: TObject);
begin
  mrMain.PlaybackMacro(synMain);
end;

procedure TfrmMain.aPrintExecute(Sender: TObject);
begin
  //TODO: Print the file :>
  //Test this ...
  if pdMain.Execute then
  begin
    printer.Title := FFileName + ' on '+DateTimeToStr(now);
    printer.BeginDoc;
    printer.Canvas.Assign(synMain.Canvas);
    printer.EndDoc;
  end;
end;

procedure TfrmMain.aRecordMacroExecute(Sender: TObject);
begin
  aRecordMacro.Checked := not aRecordMacro.Checked;
  if aRecordMacro.Checked then
    mrMain.RecordMacro(synMain)
  else
    mrMain.Stop;
end;

procedure TfrmMain.aRedoExecute(Sender: TObject);
begin
  synMain.Redo;
end;

procedure TfrmMain.aReplaceExecute(Sender: TObject);
begin
  //TODO!
end;

procedure TfrmMain.aRunExecute(Sender: TObject);
begin
  if (FFileName <> '') or (SaveFile(true)) then
  begin
    if synMain.Modified then
      SaveFile(false);
    RunFile(false);
  end;
end;

procedure TfrmMain.aSaveAsExecute(Sender: TObject);
begin
  SaveFile(true);
end;

procedure TfrmMain.aSaveExecute(Sender: TObject);
begin
  SaveFile(false);
end;

procedure TfrmMain.aSelectAllExecute(Sender: TObject);
begin
  synMain.SelectAll;
end;

procedure TfrmMain.aToggleBreakpointExecute(Sender: TObject);
begin
  ChangeBreakPoint(synMain.CaretY);
  synMain.Repaint;
end;

procedure TfrmMain.aToggleSearchExecute(Sender: TObject);
begin
  pnlSearch.Visible := not pnlSearch.Visible;
  if pnlSearch.Visible then
  begin
    if synMain.SelText <> '' then
      edSearch.Text := synMain.SelText;
    edSearch.SetFocus;
    edSearch.SelectAll;
  end else begin
    synMain.SetHighlightSearch('', []);
    synMain.SetFocus;
  end;
end;

procedure TfrmMain.aUndoExecute(Sender: TObject);
begin
  synMain.Undo;
end;

procedure TfrmMain.edSearchChange(Sender: TObject);
var AOptions: TSynSearchOptions;
begin
    AOptions := [];
    if cbMatchCase.Checked then
      AOptions := AOptions + [ssoMatchCase];
    if cbWholeWord.Checked then
      AOptions := AOptions + [ssoWholeWord];
    if cbRegExp.Checked then
      AOptions := AOptions + [ssoRegExpr];

    synMain.SetHighlightSearch(edSearch.Text, AOptions);
end;

procedure TfrmMain.edSearchKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if(key=Ord('F')) and (ssCtrl in Shift) then
  begin
    aToggleSearch.Execute;
    Key := 0;
  end;
end;

procedure TfrmMain.edSearchKeyPress(Sender: TObject; var Key: char);
var
  s, t: String;
  a, i, p: Integer;
  found, returned: Boolean;
begin
  if key = #13 then
  begin
    key := #0;
    if edSearch.Text = '' then
    begin
      aToggleSearch.Execute;
    end
    else
    begin
      a := synMain.CaretY;
      s := edSearch.Text;
      found := false;
      returned := false;
      if not cbMatchCase.Checked then
        s := AnsiLowerCase(s);

      repeat
        if returned then
          a := 1;
        for i := a-1 to synMain.Lines.Count-1 do
        begin
          t := synMain.Lines[i];
          if not cbMatchCase.Checked then
            t := AnsiLowerCase(t);
          if i = a-1 then
            p := PosEx(s, synMain.Lines[a-1], SynMain.CaretX+1)
          else
            p := Pos(s, t);
          if p > 0 then
          begin
            synMain.CaretY:=i+1;
            synMain.CaretX:=p;
            synMain.BlockBegin:=Point(p,i+1);
            synMain.BlockEnd:=Point(p+length(s),i+1);
            found := true;
            break;
          end;
        end;
        if returned then
          break
        else
          returned := true;
      until found;
    end;
    if found then
      edSearch.Color:=clWindow
    else
      edSearch.Color:=clRed;
  end;
end;

procedure TfrmMain.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  SaveSettings;
end;

procedure TfrmMain.FormCloseQuery(Sender: TObject; var CanClose: boolean);
begin
  CanClose := AskForSave;
end;

{ TLanguageList }

function TLanguageList.GetCompileString(language: String): String;
var i: Integer;
begin
  i := 0;
  result := '';
  while i < Count do
  begin
    if Strings[i] = language then
    begin
      result := Strings[i+1];
      exit;
    end;
    i := i + 5;
  end;
end;

function TLanguageList.GetRunString(language: String): String;
var i: Integer;
begin
  i := 0;
  result := '';
  while i < Count do
  begin
    if Strings[i] = language then
    begin
      result := Strings[i+2];
      exit;
    end;
    i := i + 5;
  end;
end;

function TLanguageList.GetExtensionString(language: String): String;
var i: Integer;
begin
  i := 0;
  result := '';
  while i < Count do
  begin
    if Strings[i] = language then
    begin
      result := Strings[i+3];
      exit;
    end;
    i := i + 5;
  end;

end;

function TLanguageList.GetHighlighter(language: String): TSynCustomHighlighter;
var i: Integer;
begin
  i := 0;
  result := nil;
  while i < Count do
  begin
    if Strings[i] = language then
    begin
      result := TSynCustomHighLighter(FOwner.FindComponent(Strings[i+4]));
      exit;
    end;
    i := i + 5;
  end;
end;

procedure TLanguageList.SetCompileString(language: String; s: String);
var i: Integer;
begin
  i := 0;
  while i < Count do
  begin
    if Strings[i] = language then
    begin
      Strings[i+1] := s;
      exit;
    end;
    i := i + 5;
  end;
end;

procedure TLanguageList.SetRunString(language: String; s: String);
var i: Integer;
begin
  i := 0;
  while i < Count do
  begin
    if Strings[i] = language then
    begin
      Strings[i+2] := s;
      exit;
    end;
    i := i + 5;
  end;
end;

procedure TLanguageList.SetExtensionString(language: String; s: String);
var i: Integer;
begin
  i := 0;
  while i < Count do
  begin
    if Strings[i] = language then
    begin
      Strings[i+3] := s;
      exit;
    end;
    i := i + 5;
  end;
end;

procedure TLanguageList.SetHighlighter(language: String;
  s: TSynCustomHighlighter);
var i: Integer;
begin
  i := 0;
  while i < Count do
  begin
    if Strings[i] = language then
    begin
      Strings[i+4] := s.Name;
      exit;
    end;
    i := i + 5;
  end;
end;

function TLanguageList.LanguageCount: Integer;
begin
  result := Count div 5;
end;

function TLanguageList.GetLanguage(index: Integer): String;
begin
  result := Strings[index*5];
end;

procedure TLanguageList.AddLanguage(language: String; CompileString: String;
  RunString: String; ExtensionString: String; Highlighter: TSynCustomHighlighter);
begin
  Add(language);
  Add(CompileString);
  Add(RunString);
  Add(ExtensionString);
  if Highlighter <> nil then
    Add(Highlighter.Name)
  else
    Add('');
end;

procedure TLanguageList.ExportToFile(FileName: String);
begin
  try
    SaveToFile(FileName);
  except
    on e: Exception do
    begin
      ShowMessage('Error saving file '+Filename+'! '+e.Message);
      exit;
    end;
  end;
end;

procedure TLanguageList.ImportFromFile(FileName: String);
begin
  try
    LoadFromFile(FileName);
  except
    on e: Exception do
    begin
      ShowMessage('Error opening file '+Filename+'! '+e.Message);
      exit;
    end;
  end;
end;

constructor TLanguageList.Create(Owner: TComponent);
begin
  FOwner := Owner;
end;

{ TSettings }

function TSettings.loadBool(id: String): Boolean;
begin
  result := ReadBool('settings', id, false);
end;

procedure TSettings.saveBool(id: String; value: Boolean);
begin
  WriteBool('settings', id, value);
end;

function TSettings.loadString(id: String): String;
begin
  result := ReadString('settings', id, '');
end;

procedure TSettings.saveString(id: String; value: String);
begin
  WriteString('settings', id, value);
end;

function TSettings.loadInt(id: String): Integer;
begin
  result := ReadInteger('settings', id, 0);
end;

procedure TSettings.saveInt(id: String; value: Integer);
begin
  WriteInteger('settings', id, value);
end;

function TSettings.loadFont(id: String): TFont;
var sl: TStringList;
begin
  result := nil;
  sl := TStringList.Create;
  try
    sl.Delimiter := '@';
    sl.DelimitedText := ReadString('settings', id, '');
    if sl.Count <> 6 then
      exit;
    result := TFont.Create;
    result.Name:=sl[0];
    result.Size:=StrToInt(sl[1]);
    if sl[2]='1' then
      result.Style:=result.Style+[fsBold];
    if sl[3]='1' then
      result.Style:=result.Style+[fsItalic];
    if sl[4]='1' then
      result.Style:=result.Style+[fsStrikeOut];
    if sl[5]='1' then
      result.Style:=result.Style+[fsUnderline];
  finally
    sl.Free;
  end;
end;

procedure TSettings.saveFont(id: String; value: TFont);
begin
  WriteString('settings', id, '"'+value.Name+'"'+'@'+IntToStr(value.Size)+'@'+BoolToStr(fsBold in value.Style, '1','0')+'@'+BoolToStr(fsItalic in value.Style, '1','0')+'@'+BoolToStr(fsStrikeOut in value.Style, '1', '0')+'@'+BoolToStr(fsUnderline in value.Style, '1', '0'));
end;

function TSettings.loadColor(id: String): TColor;
begin
  result := Graphics.StringToColor(ReadString('settings', id, ''));
end;

procedure TSettings.saveColor(id: String; value: TColor);
begin
  WriteString('settings', id, ColorToString(value));
end;

initialization
  {$I umain.lrs}

end.

