unit uOptions;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, Controls, Graphics, Dialogs,
  StdCtrls, Buttons, ComCtrls, EditBtn, Spin, SynEdit, CheckLst, ExtCtrls, ActnList;

type

  { TfrmOptions }

  TfrmOptions = class(TForm)
    btnCancel: TButton;
    btnOk: TButton;
    btnUp: TButton;
    btnDown: TButton;
    btnNew: TButton;
    btnColors: TButton;
    cbBorlandStyle: TCheckBox;
    cbMultibleWindows: TCheckBox;
    clbButtons: TCheckListBox;
    edCompile: TEdit;
    edExtension: TEdit;
    edDebug: TEdit;
    edRun: TEdit;
    Label1: TLabel;
    lblCompile: TLabel;
    lblRun: TLabel;
    lblExtension: TLabel;
    lblTabSize: TLabel;
    lbLanguages: TListBox;
    pcOptions: TPageControl;
    pnlButtons: TPanel;
    seTabSize: TSpinEdit;
    tsGeneral: TTabSheet;
    tsLanguages: TTabSheet;
    tsButtons: TTabSheet;
    procedure btnUpClick(Sender: TObject);
    procedure btnDownClick(Sender: TObject);
    procedure btnNewClick(Sender: TObject);
    procedure btnColorsClick(Sender: TObject);
    procedure clbButtonsClickCheck(Sender: TObject);
    procedure clbButtonsDrawItem(Control: TWinControl; Index: Integer;
      ARect: TRect; State: TOwnerDrawState);
    procedure edCompileChange(Sender: TObject);
    procedure edExtensionChange(Sender: TObject);
    procedure edRunChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure lbLanguagesSelectionChange(Sender: TObject; User: boolean);
  private
    { private declarations }
    procedure ListButtons;
    procedure ListLanguages;
  public
    { public declarations }

  end; 

var
  frmOptions: TfrmOptions;

implementation

uses uMain, uColors;

{ TfrmOptions }

procedure TfrmOptions.FormShow(Sender: TObject);
begin
  ListLanguages;
  ListButtons;
end;

procedure TfrmOptions.edCompileChange(Sender: TObject);
var s: String;
begin
  s := lbLanguages.GetSelectedText;
  frmMain.LanguageList.SetCompileString(Copy(s, 1, length(s)-1), edCompile.Text);
end;

procedure TfrmOptions.btnUpClick(Sender: TObject);
var i: Integer;
begin
  i := clbButtons.ItemIndex;
  if i >= 1 then
  begin
    frmMain.ExchangeActionItems(i,i-1);
    frmMain.CreateToolButtons;
    ListButtons;
    clbButtons.ItemIndex:=i-1;
  end;
end;

procedure TfrmOptions.btnDownClick(Sender: TObject);
var i: Integer;
begin
  i := clbButtons.ItemIndex;
  if i <= clbButtons.Count-2 then
  begin
    frmMain.ExchangeActionItems(i,i+1);
    frmMain.CreateToolButtons;
    ListButtons;
    clbButtons.ItemIndex:=i+1;
  end;
end;

procedure TfrmOptions.btnNewClick(Sender: TObject);
var s: String; ok: Boolean; i: Integer;
begin
  s:='';
  if InputQuery('Name', 'Enter the name for the new language.', s) then
  begin
    if s <> '' then
    begin
      s := AnsiLowerCase(s);
      ok := true;
      for i := 0 to frmMain.LanguageList.LanguageCount-1 do
      begin
        if s = frmMain.LanguageList.GetLanguage(i) then
          ok := false;
      end;
      if not ok then
      begin
        ShowMessage('Language does allready exist!');
      end else begin
        frmMain.LanguageList.AddLanguage(s, '', '', '', nil);

      end;
    end;
  end;
end;

procedure TfrmOptions.btnColorsClick(Sender: TObject);
begin
  frmColors.ShowModal;
end;

procedure TfrmOptions.clbButtonsClickCheck(Sender: TObject);
var i: Integer;
begin
  for i := 0 to frmMain.alMain.ActionCount-1 do
  begin
    if clbButtons.Checked[i] then
      frmMain.alMain.Actions[i].Tag := 1
    else
      frmMain.alMain.Actions[i].Tag := 0;
  end;
  frmMain.CreateToolButtons;
end;

procedure TfrmOptions.clbButtonsDrawItem(Control: TWinControl; Index: Integer;
  ARect: TRect; State: TOwnerDrawState);
var ii: Integer;
begin
  ii := TAction(frmMain.alMain.Actions[Index]).ImageIndex;
  if ii <> -1 then
  begin
    frmMain.ilMain.Draw(clbButtons.Canvas, ARect.Left, ARect.Top, ii);
  end;
end;

procedure TfrmOptions.edExtensionChange(Sender: TObject);
var s: String;
begin
  s := lbLanguages.GetSelectedText;
  frmMain.LanguageList.SetExtensionString(Copy(s, 1, length(s)-1), edExtension.Text);
end;

procedure TfrmOptions.edRunChange(Sender: TObject);
var s: String;
begin
  s := lbLanguages.GetSelectedText;
  frmMain.LanguageList.SetRunString(Copy(s, 1, length(s)-1), edRun.Text);
end;

procedure TfrmOptions.lbLanguagesSelectionChange(Sender: TObject; User: boolean);
var s: String;
begin
  s := lbLanguages.GetSelectedText;
  edCompile.Text := frmMain.LanguageList.GetCompileString(Copy(s, 1, length(s)-1));
  edRun.Text := frmMain.LanguageList.GetRunString(Copy(s, 1, length(s)-1));
  edExtension.Text := frmMain.LanguageList.GetExtensionString(Copy(s, 1, length(s)-1));
end;

procedure TfrmOptions.ListButtons;
var i: Integer; a: TAction;
begin
  clbButtons.Items.Clear;
  for i := 0 to frmMain.alMain.ActionCount-1 do
  begin
    a := TAction(frmMain.alMain.Actions[i]);
    clbButtons.Items.Add(a.Caption);
    clbButtons.Checked[i]:=frmMain.alMain.Actions[i].Tag=1;
  end;
end;

procedure TfrmOptions.ListLanguages;
var i: Integer;
begin
  lbLanguages.Clear;
  for i := 0 to frmMain.LanguageList.LanguageCount-1 do
  begin
    lbLanguages.Items.Add(frmMain.LanguageList.GetLanguage(i));
  end;
  if lbLanguages.Count>0 then
    lbLanguages.Selected[0] := true;
end;

initialization
  {$I uoptions.lrs}

end.

